unit kamera_main;

{
*********************[ Sources zu DGL Kamera Tutorial VCL Demo ]****************

        visit us at http://dgl.quellcodes.de/

 Have Fun!
     Delphic (redsoftware@gmx.de)

     Last Update 2002-10-8
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  OpenGL12, Geometry, StdCtrls, ExtCtrls, template_Fullscreen, kameras,
  OpenIL, OpenILU, OpenILUT, HiPerfTime, MMSystem, Joystick;

const
  //Entfernung der Clipping planes
  FarClipping   = 10000.0;
  NearClipping  = 1.0;

  //Hilfe Text
  HelpText : Array[0..9] of String = ('H - Zeige Hilfe',
                                      'F - Zeige Framerate + Debug Infos',
                                      'R - Zeige Renderer Info',
                                      'V - zw. Vollbild/Fenster wechseln',
                                      'L - Look Target wechseln',
                                      'Q - Look Mode wechseln',
                                      'M - Move Target wechseln',
                                      'W - Move Mode wechseln',
                                      '+/- Geschwindigkeit erhoehen/erniedrigen',
                                      'Joystick - Kameradrehung');

type
  TDemoForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
  private
    { Private-Deklarationen }
    OpenGLInitialized : Boolean; //Ist true, wenn OpenGL bereit zum zeichnen ist
    RC : HGLRC;         //OpenGL Rendering Context
    //Daten, fr den Framecounter
    StartTick : Cardinal;
    Frames : LongInt;
    //Schalter fr Hilfetexte, etc.
    ShowFPS : Boolean;
    ShowRendererInfo : Boolean;
    ShowHelp : Boolean;
    //Start der Sisplay Listen fr die Schriftwarten
    FontBase : TGLuint;
    //Debug Texte, wie z.B. FPS, etc.
    FrameText : String;
    //Funktion zum anzeigen von Text

    //Die Planeten
    Planeten : Array of TKugel;
    //Texturen
    Sun, Earth, Moon, Jupiter, Ganymede, Uranus : TGLUInt;

    //Zeiten
    FrameElapsedTime : Single;
    TimeStruct : TTimeCountStruct;
    TimeFactor : Single;

    //Kameraobjekt
    Kamera : TKamera;
    Joystick : TJoystick;

    procedure PrintText(const Text : String);
  public
    { Public-Deklarationen }
  end;

var
  DemoForm: TDemoForm;

implementation

{$R *.DFM}

procedure TDemoForm.FormCreate(Sender: TObject);
begin
  //OpenGL ist noch nicht bereit
  OpenGLInitialized := False;
  //Nur starten, wenn die Bibliotheken auch geladen sind
  if not LoadOpenGL then
    Halt(100);
  InitJoystick(Joystick, JoystickID1)
end; (*FormCreate*)

procedure TDemoForm.FormShow(Sender: TObject);
  procedure SetPixelFormat;
  var
    PixelFormat      : TGLuint;
    PFD              : pixelformatdescriptor;
  begin
    with pfd do
      begin
        nSize:= SizeOf( PIXELFORMATDESCRIPTOR );
        nVersion:= 1;
        dwFlags:= PFD_DRAW_TO_WINDOW
          or PFD_SUPPORT_OPENGL
          or PFD_DOUBLEBUFFER;
        iPixelType:= PFD_TYPE_RGBA;
        cColorBits:= 16;             //Farbtiefe
        cRedBits:= 0;
        cRedShift:= 0;
        cGreenBits:= 0;
        cBlueBits:= 0;
        cBlueShift:= 0;
        cAlphaBits:= 0;
        cAlphaShift:= 0;
        cAccumBits:= 0;              //Accumulation Buffer
        cAccumRedBits:= 0;
        cAccumGreenBits:= 0;
        cAccumBlueBits:= 0;
        cAccumAlphaBits:= 0;
        cDepthBits:= 16;             //Z-Buffer Tiefe
        cStencilBits:= 0;            //Stencil Buffer
        cAuxBuffers:= 0;
        iLayerType:= PFD_MAIN_PLANE;
        bReserved:= 0;
        dwLayerMask:= 0;
        dwVisibleMask:= 0;
        dwDamageMask:= 0
      end;
    //Pixel Format setzten
    PixelFormat := ChoosePixelFormat(Canvas.Handle, @pfd);
    if (PixelFormat=0) then
      MessageDlg('Es konnte kein passendes PixelFormat gefunden werden.', mtError, [mbOk], 0);
    if (not Windows.SetPixelFormat(Canvas.Handle,PixelFormat,@pfd)) then
      MessageDlg('PixelFormat konnte nicht gesetzt werden.', mtError, [mbOk], 0)
  end; (*SetPixelFormat*)

  procedure StartGl;
  begin
    //Rendering Context initialisieren
    RC := wglCreateContext(Canvas.Handle);
    if (RC=0) then
    begin
      MessageDlg('Rendering Context kann nicht erstellt werden.', mtError, [mbOk], 0);
      Halt(100)
    end;

    if (not wglMakeCurrent(Canvas.Handle, RC)) then
    begin
      MessageDlg('Rendering Context kann nicht aktiviert werden.', mtError, [mbOk], 0);
      Halt(100)
    end
  end; (*StartGl*)

  procedure SetupGL;
  begin
    //OpenGL Voreinstellungen
    glClearColor(0.3, 0.3, 0.3, 0.0); //Hintergrundfarbe
    glEnable(GL_DEPTH_TEST); //Tiefenpuffer aktivieren
    glEnable(GL_CULL_FACE); //Backface Culling aktivieren
    //glPolygonMode(GL_BACK, GL_LINE);

    glShadeModel(GL_SMOOTH)
  end; (*SetupGL*)

  procedure BuildFont;
  var
    font : HFONT;
  begin
    //Schriftart erstellen
    FontBase := glGenLists(96);

    font := CreateFont(16,                // height of font
         0,                             // average character width
         0,                             // angle of escapement
         0,                             // base-line orientation angle
         FW_MEDIUM,                     // font weight
         0,                             // italic
         0,                             // underline
         0,                             // strikeout
         ANSI_CHARSET,                  // character set
         OUT_TT_PRECIS,                 // output precision
         CLIP_DEFAULT_PRECIS,           // clipping precision
         ANTIALIASED_QUALITY,           // output quality
         FF_DONTCARE or DEFAULT_PITCH,  // pitch and family
         'Courier');                    // font
    SelectObject(Canvas.Handle, font);

    wglUseFontBitmaps(Canvas.Handle, 32, 96, FontBase);

    DeleteObject(font)
  end; (*Build Font*)

  procedure LoadTextures;
    function LoadDevILTexture(Filename : String): TGLUInt;
    var
      id:TILInt;
      imageInfo : TILInfo;
    begin
      {$T-}
      // Laden eines Texture in den Speicher
      //if ilutEnable(ILUT_GL_USE_S3TC) = ILUT_INVALID_ENUM then
      //  Exit;
      ilGenImages(1, @id);
      ilBindImage(id);
      ilLoadImage(PChar(Filename));  // Loads into the current bound image

      iluGetImageInfo(@ImageInfo);

      // bergeben der Texture an OpenGL
      Result := ilutGLBindMipmaps;
      glBindTexture(GL_TEXTURE_2D, Result);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

      // Da OpenGL die Texture hat, knnen wir sie aus dem Speicher entfernen
      ilDeleteImages(1, @Id);
      {$T+}
    end;

  begin
    ilInit;
    ilutRenderer(IL_OPENGL);
    glEnable(GL_TEXTURE_2D);
    Sun := LoadDevILTexture('textures\sun.jpg');
    Earth := LoadDevILTexture('textures\earth.jpg');
    Moon := LoadDevILTexture('textures\moon.jpg');
    Jupiter := LoadDevILTexture('textures\jupiter.jpg');
    Ganymede := LoadDevILTexture('textures\ganymede.jpg');
    Uranus := LoadDevILTexture('textures\uranus.jpg');
  end;


  procedure InitPlaneten;
  begin
    //Planeten initialisieren
    SetLength(Planeten, 6);
    //Sonne
    Planeten[0] := TKugel.Create;
    with Planeten[0] do
    begin
      Objektname := 'Sonne';
      TextureID := Sun;
      Groesse := 10.0;
      RotVel := 10;
      CreateDisplayList
    end;

    //Erde
    Planeten[1] := TAttachedKugel.Create;
    with Planeten[1] as TAttachedKugel do
    begin
      Objektname := 'Erde';
      TextureID := Earth;
      Groesse := 1.0;
      RotVel := 90;

      Entfernung := 20.0;
      Umlaufdauer := 20;
      Parent := Planeten[0];
      CreateDisplayList
    end;

    //Mond
    Planeten[2] := TAttachedKugel.Create;
    with Planeten[2] as TAttachedKugel do
    begin
      Objektname := 'Mond';
      TextureID := Moon;
      Groesse := 0.237;
      RotVel := -180;

      Entfernung := 2.5;
      Umlaufdauer := 8;

      Parent := Planeten[1];
      CreateDisplayList
    end;

    //Jupiter
    Planeten[3] := TAttachedKugel.Create;
    with Planeten[3] as TAttachedKugel do
    begin
      Objektname := 'Jupiter';
      TextureID := Jupiter;
      Groesse := 11.2;
      RotVel := -60;

      Entfernung := 100;
      Umlaufdauer := 60;

      Parent := Planeten[0];
      CreateDisplayList
    end;

    //Ganymed
    Planeten[4] := TAttachedKugel.Create;
    with Planeten[4] as TAttachedKugel do
    begin
      Objektname := 'Ganymed';
      TextureID := Ganymede;
      Groesse := 1;
      RotVel := -200;

      Entfernung := 15;
      Umlaufdauer := 7.2;

      Parent := Planeten[3];
      CreateDisplayList
    end;

    //Uranus
    Planeten[5] := TAttachedKugel.Create;
    with Planeten[5] as TAttachedKugel do
    begin
      Objektname := 'Uranus';
      TextureID := Uranus;
      Groesse := 4.1;
      RotVel := 360;

      Entfernung := 150;
      Umlaufdauer := 120;

      Parent := Planeten[0];
      CreateDisplayList
    end;

  end;

begin
  RC := 0;

  SetPixelFormat;
  StartGL;

  //GL Extensions neu auslesen
  ClearExtensions;
  ReadExtensions;

  SetupGL;
  BuildFont;
  LoadTextures;
  InitLighting;
  InitPlaneten;

  FrameElapsedTime := 0.001;
  BeginCounting(TimeStruct);
  TimeFactor := 1.0/1.0;

  //Kamera initialisieren
  Kamera := TKamera.Create;
  Kamera.MoveMode := mmAttatched;
  Kamera.MoveTarget := Planeten[2];
  Kamera.LookMode := lmLookAtObjekt;
  Kamera.LookTarget := Planeten[0];

  //Kleinkram initialisieren
  Frames := 0;
  StartTick := GetTickCount;
  ShowFPS := True;
  ShowHelp := True;
  ShowRendererInfo := False;

  //OnIdle setzen
  Application.OnIdle := ApplicationEventsIdle;

  //OpenGL ist nun bereit
  OpenGLInitialized := True
end; (*FormShow*)

procedure TDemoForm.FormDestroy(Sender: TObject);
var
  I : Integer;
begin
  //Shcriftart wieder freigeben
  glDeleteLists(FontBase, 96);

  for I := Low(Planeten) to High(Planeten) do
    Planeten[I].Free;

  OpenGLInitialized := False;
  if RC<>0 then
    begin
      if (not wglMakeCurrent(Canvas.Handle,0)) then
        MessageDlg('Rendering Context konnte nicht abgewhlt werden.', mtError, [mbOk], 0);
      if (not wglDeleteContext(RC)) then
        MessageDlg('Rendering Context konnte nicht freigegeben werden.', mtError, [mbOk], 0)
    end;
  RC:=0
end; (*Form Destroy*)

procedure TDemoForm.FormResize(Sender: TObject);
begin
  glViewport(0, 0, ClientWidth, ClientHeight);
  //Projektionsmatrix resetten
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //Perspektivische Darstellung
  gluPerspective(60.0,ClientWidth/ClientHeight,NearClipping,FarClipping);// Calculate The Aspect Ratio Of The Window
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity
end; (*FormResize*)

procedure TDemoForm.ApplicationEventsIdle(Sender: TObject;
  var Done: Boolean);

  procedure ShowText;
  var
    Line : LongInt;

    procedure LineOut(const X : LongInt; Text : String);
    begin
      glRasterPos2i(X, (Line)*16);
      PrintText(Text);
      Inc(Line)
    end;(*LineOut*)

  const
    cLookMode : Array[TLookMode] of String = ('lmFree', 'lmObjektSight', 'lmLookAtObjekt');
    cMoveMode : Array[TMoveMode] of String = ('mmFree', 'mmAttatched');

  var
    I : Integer;
  begin
    //Text ausgabe
    Line := 1;
    glPushAttrib(GL_DEPTH_BUFFER_BIT or GL_ENABLE_BIT);
      glDisable(GL_DEPTH_TEST);
      glDisable(GL_LIGHTING);
      glDisable(GL_TEXTURE_2D);

      //Wir brauchen eine Orthogonale ansicht fr Text

        // Aktuelle Projektionsmatrix auf den Stack
      glMatrixMode(GL_PROJECTION);
        glPushMatrix();
          glLoadIdentity();

          //Ortho Matrix
          glOrtho(0.0, ClientWidth, ClientHeight, 0, -1, 1);

          // Aktuelle Modelviewmatrix auf den Stack
          glMatrixMode(GL_MODELVIEW);
          glPushMatrix();
            glLoadIdentity();

            // Text Farbe
            glColor3f(0.0, 1.0, 1.0);

            //Hilfe Text ausgeben, wenn angeschaltet
            if ShowHelp then
            begin
              LineOut(5, 'Hilfe:');
              for i := Low(HelpText) to High(HelpText) do
                LineOut(5, HelpText[i]);
              LineOut(5, '')
            end;

            //Renderer Infos ausgeben, wenn angeschaltet
            if ShowRendererInfo then
            begin
              LineOut(5, 'Info:');
              LineOut(5, 'Renderer: '+ glGetString(GL_RENDERER));
              LineOut(5, 'Vendor: '+ glGetString(GL_VENDOR));
              LineOut(5, 'Version: '+ glGetString(GL_VERSION));
              LineOut(5, '')
            end;

            //Kamera infos ausgeben
            LineOut(5, 'Kamera:');
            LineOut(5, 'Look Mode: ' + cLookMode[Kamera.LookMode]);
            LineOut(5, '  Look Target: "'+ Kamera.LookTarget.ObjektName + '"');
            LineOut(5, 'Move Mode: ' + cMoveMode[Kamera.MoveMode]);
            LineOut(5, '  Move Target: "'+ Kamera.MoveTarget.ObjektName + '"');

            //FrameText(FrameRate) ausgeben
            if ShowFPS then
            begin
              glColor3f(1.0, 0.0, 0.0);
              glRasterPos2i(5, ClientHeight-16);
              PrintText(FrameText)
            end;

          //Projektionsmatrix wiederherstellen
          glMatrixMode(GL_PROJECTION);
          glPopMatrix();

        //Modelview Matrix wiederherstellen
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();

    glPopAttrib()
  end; (*ShowText*)

  procedure Billboard;
  const
    NewModelView : Array[0..11] of TGLFloat =
      (1,0,0,0,0,1,0,0,0,0,1,0);
  var
    Modelview : Array[0..15] of TGLFloat;

  begin
    // Modelview Matrix holen
    glGetFloatv(GL_MODELVIEW_MATRIX , @Modelview[0]);
    Move(NewModelView[0], ModelView[0], SizeOf(NewModelview));
    glLoadMatrixf(@Modelview[0])
  end;

var
  Error : LongInt; //Fehler?
  I : Integer;

begin
  //Abbrechen, wenn OpenGL noch nicht initialisiert ist
  if not OpenGLInitialized then Exit;
  //Immer schn neu zeichnen
  Done := False;

  //Puffer killen
  glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);

  //Planeten bewegen
  for I := Low(Planeten) to High(Planeten) do
    Planeten[I].Move(FrameElapsedTime*TimeFactor);

  PollJoystick(Joystick);
  Kamera.UpdateRotation(JoyStick, FrameElapsedTime);

  //Kamera initialisieren
  glLoadIdentity;
  Kamera.UseMatrix;

  //Planeten rendern
  for I := Low(Planeten) to High(Planeten) do
    Planeten[I].Render;

  {with Planeten[3] do
    glTranslatef(Position[0], Position[1], Position[2]);

  Billboard;

  glBegin(GL_QUADS);
    glVertex3f(-10, -10, 0);
    glVertex3f(10, -10, 0);
    glVertex3f(10, 10, 0);
    glVertex3f(-10, 10, 0);
  glEnd();}

  //Text anzeigen
  ShowText;

  //Error Handler
  Error := glgetError;
  if Error <> GL_NO_ERROR then
  begin
    Caption := gluErrorString(Error);
    //Rendering kurz anhalten
    Done := True;
    FlashWindow(Handle, True)
  end;
  //Frame Counter
  Inc(Frames);
  if GetTickCount - StartTick >=500 then
  begin
    FrameText := Format('FPS: %f Zeitraffer:%f', [Frames/(GetTickCount - StartTick)*1000, 1/TimeFactor]);
    Frames := 0;
    StartTick := GetTickCount
  end;
  //Und anzeigen
  SwapBuffers(Canvas.Handle);
  FrameElapsedTime := GetTimeNeeded(TimeStruct);
  BeginCounting(TimeStruct)
end; (*ApplicationEventsIdle*)

procedure TDemoForm.FormKeyPress(Sender: TObject; var Key: Char);
var
  i : LongInt;
begin
  //Tasten krams
  case UpCase(Key) of
    'H': ShowHelp := not ShowHelp;
    'F': ShowFPS := not ShowFPS;
    'R': ShowRendererInfo := not ShowRendererInfo;
    'V': begin
           SwichFullscreenMode;
         end;
    'L': begin
           i := 0;
           while Kamera.LookTarget <> Planeten[I] do
             Inc(i);
           Inc(i);
           if i > High(Planeten) then i := Low(Planeten);
           Kamera.LookTarget := Planeten[i]
         end;
    'M': begin
           i := 0;
           while Kamera.MoveTarget <> Planeten[I] do
             Inc(i);
           Inc(i);
           if i > High(Planeten) then i := Low(Planeten);
           Kamera.MoveTarget := Planeten[i]
         end;
    'W' : begin
            if Kamera.MoveMode = High(TMoveMode) then
              Kamera.MoveMode := Low(TMoveMode)
            else
              Inc(Kamera.MoveMode)
          end;
    'Q' : begin
            if Kamera.LookMode = High(TLookMode) then
              Kamera.LookMode := Low(TLookMode)
            else
              Inc(Kamera.LookMode)
          end;
    'A' : begin
            if Kamera.LookMode = Low(TLookMode) then
              Kamera.LookMode := High(TLookMode)
            else
              Dec(Kamera.LookMode)
          end;
    '+' : TimeFactor := TimeFactor * 2.0;
    '-' : TimeFactor := TimeFactor / 2.0;
  end
end; (*FormKeyPress*)

{$Q-}
procedure TDemoForm.PrintText(const Text: String);
begin
  //Zeigt Text an Rasterpostion an, berlaufchecks mssen aber deaktiviert sein($Q-}
  if (Text = '') then
    Exit;
  glPushAttrib(GL_LIST_BIT);
    glListBase(FontBase - 32);  // Set the base list to our character list
    glCallLists(Length(text), GL_UNSIGNED_BYTE, PChar(Text)); // Display the text
  glPopAttrib()
end; (*PrintText*)
{$Q+}

end.


