unit kameras;

{
**********************[  Sources zu DGL Kamera Tutorial  ]**********************

        visit us at http://dgl.thechaoscompany.net/

 Have Fun!
     Delphic (redsoftware@gmx.de)

     Last Update 2002-10-08
}

interface
uses
  OpenGl12, Geometry, Joystick;

type
  //Basis Objekt
  TObjekt = class
  private
  public
    ObjektName : String;
    Rotation  : THomogeneousMatrix;
    Position  : TVertex;
    constructor Create; virtual;
    procedure UpdateRotation(const Joystick : TJoyStick; Time : Single);
  end;

  //Planeten
  TKugel = class(TObjekt)
  private
  public
    TextureID : TGlUInt;
    ListID    : TGlUInt;
    Groesse   : Single;
    RotVel : Single;

    constructor Create; override;
    destructor Destroy; override;
    procedure CreateDisplayList;
    procedure Render;
    procedure Move(Time : Single); virtual;
  end;

  //Angehngter Planet
  TAttachedKugel = class(TKugel)
  private
    Umlaufposition : Single;
  public
    Umlaufdauer : Single;
    Entfernung  : Single;
    Parent      : TKugel;
    procedure Move(Time : Single); override;
    constructor Create; override;
  end;

  //Alles fr die Kamera
  TLookMode = (lmFree, lmObjektSight, lmLookAtObjekt);
  TMoveMode = (mmFree, mmAttatched);
  TKamera = class(TObjekt)
  private
  public
    LookMode : TLookMode; //Ansichtsmodus
    MoveMode : TMoveMode; //Bewegungsmodus

    LookTarget, MoveTarget : TObjekt; //Ansichtsziel und Bewegungsziel

    procedure UseMatrix; virtual; //Kamera-Matrix einsetzen
    constructor Create; override;
    procedure ResetView; //Ansicht zurcksetzen
  end;

procedure InitLighting;

implementation

procedure InitLighting;
const
  light_position : array[0..3] of TGLfloat = (0.0, 0.0, 500.0, 0.0);
  mat_specular  : array[0..3] of TGLfloat = (0.3, 0.3, 0.3, 1.0);
  mat_shininess : array[0..0] of TGLfloat = (15.0);

  mat_diffuse : array[0..3] of TGLfloat = (0.7, 0.7, 0.7, 0.0);
begin
  glEnable(GL_LIGHTING);
  glMaterialfv(GL_FRONT, GL_SPECULAR,  @mat_specular[0]);
  glMaterialfv(GL_FRONT, GL_SHININESS, @mat_shininess[0]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE,   @mat_diffuse[0]);

  glLightfv(GL_LIGHT0, GL_POSITION, @light_position[0]);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0)
end;

{ TKugel }

constructor TKugel.Create;
begin
  inherited;
  TextureID := 0;
  ListID := 0;
  Groesse := 1.0;

  RotVel := 90
end;

procedure TKugel.CreateDisplayList;
var
  qObj : PGLUQuadric;
begin
  if (ListID <> 0) then glDeleteLists(ListID, 1);
  ListID := glGenLists(1);

  qObj := gluNewQuadric();

  gluQuadricDrawStyle(qObj, GLU_FILL);
  gluQuadricNormals(qObj, GLU_SMOOTH);
  gluQuadricTexture(qObj, True);

  glNewList(ListID, GL_COMPILE);
    gluSphere(qObj, Groesse, 32, 32);
  glEndList();

  gluDeleteQuadric(qObj)
end;

destructor TKugel.Destroy;
begin
  if (ListID <> 0) then glDeleteLists(ListID, 1);
  ListID := 0;
  Inherited
end;

procedure TKugel.Move(Time: Single);
begin
  Rotation := MatrixMultiply(Rotation, CreateRotationMatrixY(Sin(Pi/180* RotVel *Time),
                                                 Cos(Pi/180* RotVel *Time)))
end;

procedure TKugel.Render;
begin
  if (ListID = 0) then Exit;
  glPushMatrix;
    glTranslatef(Position[0], Position[1], Position[2]);
    glMultMatrixf(@Rotation[0,0]);
    glBindTexture(GL_TEXTURE_2D, TextureID);
    glCallList(ListID);
  glPopMatrix
end;

{ TAtachedKugel }

constructor TAttachedKugel.Create;
begin
  inherited;
  Umlaufdauer := 10;
  Entfernung := 10;
  Umlaufposition := 0;
  Parent := nil;
end;

procedure TAttachedKugel.Move(Time: Single);
var
  Verschiebung : TVertex;
begin
  inherited;

  if Assigned(Parent) then
    Position := Parent.Position
  else
  begin
    Position[0] := 0.0;
    Position[1] := 0.0;
    Position[2] := 0.0;
  end;

  Umlaufposition := Umlaufposition + 2* Pi * Time / Umlaufdauer;
  Verschiebung[0] := Sin(Umlaufposition)*Entfernung;
  Verschiebung[1] := 0;
  Verschiebung[2] := Cos(Umlaufposition)*Entfernung;

  //Achtung, das ist hier nur reingeflickt, weil ich zu faul war
  //eine extra Klasse fr den Mond zu erstellen. Bitte
  if ObjektName = 'Mond' then
    Verschiebung[1] := Sin(Umlaufposition)*0.3;

  Position := VectorAdd(Position, Verschiebung)
end;

{ TObjekt }

constructor TObjekt.Create;
begin
  Rotation := IdentityHmgMatrix;
  ObjektName := '<unnamed>';

  Position[0] := 0.0;
  Position[1] := 0.0;
  Position[2] := 0.0
end;

procedure TObjekt.UpdateRotation(const Joystick: TJoyStick; Time : Single);
var
  XAngle, YAngle : Single;
begin
  if ((Joystick.y - 0.5) > 0.1) or ((Joystick.y - 0.5) < -0.1) then
    YAngle := 2 * Pi / 2 * (Joystick.Y - 0.5) * Time
  else
    YAngle := 0;
  Rotation := MatrixMultiply(CreateRotationMatrixX(Sin(YAngle), Cos(YAngle)),Rotation);

  if ((Joystick.x - 0.5) > 0.1) or ((Joystick.x - 0.5) < -0.1) then
    XAngle := -2 * Pi / 2 * (Joystick.x - 0.5) * Time
  else
    XAngle := 0;

  if JoyStick.Buttons[1] then
    Rotation := MatrixMultiply(CreateRotationMatrixZ(Sin(XAngle), Cos(XAngle)),Rotation)
  else
    Rotation := MatrixMultiply(CreateRotationMatrixY(Sin(XAngle), Cos(XAngle)),Rotation)
end;

{ TKamera }

constructor TKamera.Create;
begin
  inherited;
  LookMode := lmFree;
  MoveMode := mmFree;

  LookTarget := nil;
  MoveTarget := nil;

end;

procedure TKamera.ResetView;
begin
  Rotation := IdentityHmgMatrix
end;

procedure TKamera.UseMatrix;
var
  RotMatrix : THomogeneousMatrix;
  Pos : TVertex;

  procedure LockTarget(Target : TObjekt);
  var
    InvMatrix, M1, M2 : THomogeneousMatrix;
    ObjektPosition : TVertex;

    Cos, Sin : Single;
    xzEbenenPosition : TVertex;
  begin
    try
      InvMatrix := Rotation;
      InvertMatrix(InvMatrix);

      //Globale Objekt Position
      ObjektPosition := VectorSubtract(Position, Target.Position);
      //In lokale Koordinaten drehen
      ObjektPosition := VectorTransform(ObjektPosition, InvMatrix);

      xzEbenenPosition[0] := ObjektPosition[0];
      xzEbenenPosition[1] := 0;
      xzEbenenPosition[2] := ObjektPosition[2];
      if not((xzEbenenPosition[0] = 0.0) and (xzEbenenPosition[2] = 0.0)) then
      begin
        NormalizeVector(xzEbenenPosition);
        M1 := CreateRotationMatrixY(xzEbenenPosition[0], xzEbenenPosition[2])
      end
      else
        M1 := IdentityHmgMatrix;

      NormalizeVector(ObjektPosition);
      Cos := VectorDotProduct(xzEbenenPosition, ObjektPosition);
      Sin := Sqrt(1- Cos*Cos);

      if (ObjektPosition[1] > 0.0) then
        Sin := - Sin;
      M2 := CreateRotationMatrixX(Sin, Cos);

      RotMatrix := MatrixMultiply(M1, M2);
      Rotation := MatrixMultiply(RotMatrix, Rotation);
      RotMatrix := Rotation;
      InvertMatrix(RotMatrix)
    except
      RotMatrix := IdentityHmgMatrix
    end
  end;

begin
  //Ansichten
  case LookMode of
    lmFree     : begin
                   RotMatrix := Rotation;
                   InvertMatrix(RotMatrix)
                 end;
    lmObjektSight: begin
                     Assert(Assigned(LookTarget));
                     Rotation := LookTarget.Rotation;
                     RotMatrix := Rotation;
                     InvertMatrix(RotMatrix)
                   end;
    lmLookAtObjekt : ; //Kommt erst nach MoveMode
  end;


  //Bewegungen
  case MoveMode of
    mmFree     : begin
                   Pos := Position
                 end;
    mmAttatched : begin
                   Assert(Assigned(MoveTarget));
                   Position := MoveTarget.Position;
                   Pos := Position;
                 end;
  end;

  if (LookMode = lmLookAtObjekt) then
  begin
    Assert(Assigned(LookTarget));
    if not ((MoveMode = mmAttatched) and (MoveTarget = LookTarget)) then
      LockTarget(LookTarget)
  end;

  //Matrix einlegen
  glLoadMatrixf(@RotMatrix[0,0]);
  glTranslatef(-Pos[0], -Pos[1], -Pos[2])
end;

end.
